/***************************************************************************
 * LPRng - An Extended Print Spooler System
 *
 * Copyright 1988-1995 Patrick Powell, San Diego State University
 *     papowell@sdsu.edu
 * See LICENSE for conditions of use.
 *
 ***************************************************************************
 * MODULE: tcp-lp.c
 * PURPOSE: open a socket to the remote printer
 **************************************************************************/

static char _id[] = "annex-lp.c,v 1.1 1997/01/05 04:54:10 papowell Exp";

/*
 * annex-lp: a PLP LP-pipe command for communicating with printers
 *		   connected to Annex terminal servers.
 *
 * You are safe using the first few arguments to communicate LP-pipe-
 * specific configuration (such as the remote hostname and remote
 * port); PLP will append it's arguments to those specified in the
 * printcap entry.
 *						      -- Justin Mason 26 Aug 94.
 *
 * (from original source file:)
 * >
 * > Open a socket to an Annex terminal server. This code has been excerpted
 * > from modifications made to the Berkeley LPD code, by ENCORE Computers Inc.,
 * > and has been used with their permission.
 * >
 * > It appears that the terminal server accepts a port connection on the LPD
 * > port. I suppose that it then treats the selected printer nicely. Not bad,
 * > as far as hacks go, but it does expose a severe security loophole.
 * >
 * > Fri Sep 30 09:59:59 CDT 1988 Patrick Powell
 *
 * lppipename annexhost line \  (these two are in the printcap entry)
 *
 * -PPrinter -wwidth -llength -xwidth -ylength [-c] [-iindent] \
 * [-Zoptions] [-Cclass] [-Jjob] [-Raccntname] -nlogin -hhost  \
 * -Fformat -Ddebug -sStatus_file [affile]
 *		      (all added by lpd)
 */

#include "lp-pipe.h"

#define OPEN_RETRY_DELAY  10      /* delay when waiting for open */

int sockfd = -1;
int wait_til_ready;
char *Host;
int Port;
int Line;
char *Name;
char *Status_file;
char buf[1024];		/* don't do write any longer than this... sigh */

void usage (void );
void annex_pipe( void );

/***************************************************************************
 * main() - parse the various options
 * - note - we want only the -s statusfile
 *
 ***************************************************************************/
int main (int argc, char **argv )
{
	int i;
	char *s;

	/*
	 * first two args are the host/port pair.
	 */
	i = 1;
	Name = argv[0];
	if( argc - i < 2 ) usage ();
	Host = argv[i++];
	if( (Line = atoi( argv[i++] )) <= 0) usage ();

	/* get the -sStatusFile argument */
	while( i < argc ){
		s = argv[i++];
		if( s[0] == '-' && s[1] == 's' ){
			Status_file = s+2;
		}
	}
	(void) signal(SIGINT, kill_job);
	(void) signal(SIGPIPE, SIG_IGN);

	annex_pipe();
	exit (JSUCC);
}

/* the current job has been lprm'ed */
plp_signal_t kill_job (int signal)
{
	/* exit with a success code; we don't want the queue to stop */
	exit (JFAIL);
}

void usage (void ){
	fprintf (stderr, "usage: tcp-lp [-W] host port [... args]\n");
	exit(JABORT);
}

#include <setjmp.h>

static jmp_buf ack_jmp_buf;
/*
 * SIGALRM catcher for Annex ACK listener.
 * If a SIGALRM comes, then we didn't get an
 * initial ACK, meaning that this Annex is probably using the
 * old Annex LPD protocol.
 */
plp_signal_t
noannexack ()
{
	longjmp (ack_jmp_buf, 1);
}

/*
 * Open a socket to an Annex terminal server.
 * This code has been excerpted from
 * modifications made to the Berkeley LPD code by ENCORE Computers Inc.,
 * and has been * used with their permission.
 * 
 * It appears that the terminal server accepts a port connection on the
 * LPD port.  I suppose that it then treats the selected printer nicely.
 * 
 * Fri Sep 30 09:59:59 CDT 1988 Patrick Powell
 */
#define SERIAL  1
#define PAR     2
#define ANNEX_SPOOL   '\011'
#define ACK_WAIT 10			 /* seconds to wait for annex to ACK */

void Open_annex()
{
	int status;				/*  even used by ENCORE */
	int n;					  /* number chars read/written */
	int err;				    /* save value of errno */
	struct servent *srv;
	fd_set rmask;				/* wait for stuff */
	struct timeval delay;   /* timeout if needed */


	setstatus ("opening Annex %s port %d", Host, Line);
	if ((srv = getservbyname ("printer", "tcp")) == NULL) {
		perror ("get port (printer)");
		Port = 515;
	} else {
		Port = srv->s_port;
	}
	sockfd = -1;
	while( sockfd < 0 ){
		/* open socket and see what happens */
		open_sock( OPEN_RETRY_DELAY );
		if( sockfd >= 0 ){
			/*
			 * you are trying to connect to the ANNEX PRINTER port,
			 *	and set up a simple transfer protocol.
			 */
			(void) sprintf (buf, "%c%d;%d\n",
				ANNEX_SPOOL,
				Line <= 0 ? PAR : SERIAL,
				Line <= 0 ? 1 : Line);
			n = strlen (buf);
			/*
			 * write a string to the terminal server
			 */
			if( BPwrite (sockfd, buf, n) > 0 ){
				delay.tv_sec = 10;
				delay.tv_usec = 0;
				FD_ZERO (&rmask);
				FD_SET (sockfd, &rmask);		/* socket */
				/* Find out where input is available */
				status = select( sockfd + 1, &rmask, 0, 0, &delay);
				err = errno;
				switch( status ){
				case -1:
					fprintf( stderr,
						"annex open: select error from %s port %d: %s\n",
							Host, Line, Errormsg(err) );
					break;
				case 0:
					fprintf( stderr,
						"annex open: ACK timeout from %s port %d: %s\n",
							Host, Line, Errormsg(err) );
					break;
				default:
					n = read( sockfd, buf, 1);
					if (n != 1) {
						fprintf( stderr,
							"annex open: read error from %s port %d: %s\n",
							Host, Line, Errormsg(err) );
					} else if( buf[0] != '\0') {
						fprintf (stderr,
						 "annex open: bad ack '%d' read from %s port %d\n",
							 buf[0], Host, Line );
					} else {
						/* success */
						/* this should exit from the loop */
						continue;
					}
				}
			}
			if( sockfd >= 0 ){
				(void) close (sockfd);
				sockfd = -1;
			}
			sleep( OPEN_RETRY_DELAY );
		}
	}
}

/***************************************************************************
 * Annex_ack()
 * (called on close)
 * 1. send an OOB ACK to indicate that you really have finished (EOF)
 * 2. Write a 0 byte to flush the output message (4.2 Bugger)
 * 3. Wait for a response that it has been done
 ***************************************************************************/
void
Annex_ack ()
{
	int n;
	int err;
	char nullstr[2];
#define ACK_CL_WAIT (10*60)

	(void) signal (SIGALRM, noannexack);
	if (setjmp (ack_jmp_buf)) {
		fprintf (stderr, "%s port %d closing ack timed out\n", Host, Line);

	} else {
		nullstr[0] = 0;
		(void) alarm (ACK_CL_WAIT);
		if( send (sockfd, nullstr, 1, MSG_OOB) != 1) {
			err = errno;
			(void) alarm (0);   /* disarm */
			fprintf (stderr, "closing ack: send-oob failed - %s",
				Errormsg(err));
			return;
		}
		if (send (sockfd, nullstr, 1, 0) != 1) {
			err = errno;
			(void) alarm (0);   /* disarm */
			fprintf (stderr, "closing ack: send failed - %s",
				Errormsg(err));
			return;
		}
		n = read (sockfd, nullstr, 1);
		err = errno;
		(void) alarm (0);	   /* disarm */

		if (n != 1) {
			fprintf (stderr, "closing ack: read error from %s port %d - %s",
				Host, Line, Errormsg(err));
		} else if (nullstr[0] != '\0') {
			fprintf (stderr,
				 "Annex_ack: '%d' read from %s port %d\n", nullstr[0], Host,
				 Line);
		}
	}
}

void annex_pipe()
{
	int in, out;
	int err;

	Open_annex();
	while ((in = read (0, buf, sizeof (buf))) > 0) {
		out = BPwrite (sockfd, buf, in);
		err = errno;
		if( out < 0 ) {
			fprintf( stderr,
				"annex-pipe: write error to %s port %d: %s\n",
					Host, Line, Errormsg(err) );
			exit (JFAIL);
		}
	}
	err = errno;
	if( in < 0 ){
		fprintf( stderr,
			"annex-pipe: read error from stdin - %s\n",
				Errormsg(err) );
		exit (JFAIL);
	}
	Annex_ack();
	(void)close (sockfd);
}
