/*
 * lp-pipe.c: PLP LP-pipe miscellaneous functions.
 *
 *                              -- Justin Mason 26 Aug 94.
 */

#include "lp-pipe.h"

/* setstatus: put a message in the "status" file, if available. */

/* VARARGS2 */
void
#ifdef __STDC__
setstatus (char *msg,...)
#else
setstatus (va_alist) va_dcl
#endif
{
#ifndef __STDC__
	char *msg;
#endif
	static FILE *fp;
	static int first_call;
	int fd;
	int err;
	VA_LOCAL_DECL

	if( first_call == 0 ){
		if( Status_file && *Status_file){
			fd = open( Status_file, O_WRONLY | O_APPEND );
			err = errno;
			if( fd < 0 ){
				fprintf(stderr, "couldn't open '%s': %s\n",
					Status_file, Errormsg(err) );
			} else if( (fp = fdopen( fd, "a" )) == 0 ){
				fprintf(stderr, "fopen failed '%s': %s\n",
					Status_file, Errormsg(err) );
			}
		}
		first_call = 1;
	}
	if( fp ){
		VA_START (msg);
		VA_SHIFT (msg, char *);
		(void) vfprintf( fp, msg, ap);
		fflush( fp );
		VA_END;
	}
}

/****************************************************************************/
#if !defined(HAVE_STRERROR)

# if defined(HAVE_SYS_NERR)
#  if !defined(HAVE_SYS_NERR_DEF)
	 extern int sys_nerr;
#  endif
#  define num_errors    (sys_nerr)
# else
#  define num_errors    (-1)            /* always use "errno=%d" */
# endif

# if defined(HAVE_SYS_ERRLIST)
#  if !defined(HAVE_SYS_ERRLIST_DEF)
	 extern const char *const sys_errlist[];
#  endif
# else
#  undef  num_errors
#  define num_errors   (-1)            /* always use "errno=%d" */
# endif

#endif

const char * Errormsg ( int err )
{
	const char *cp;

#if defined(HAVE_STRERROR)
	cp = strerror(err);
#else
# if defined(HAVE_SYS_ERRLIST)
	if (err >= 0 && err < num_errors) {
		cp = sys_errlist[err];
	} else
# endif
	{
		static char msgbuf[32];     /* holds "errno=%d". */
		/* SAFE use of sprintf */
		(void) sprintf (msgbuf, "errno=%d", err);
		cp = msgbuf;
	}
#endif
	return (cp);
}

/***************************************************************************
 * open_sock()
 * - open a socket to the remote host/port
 *  if retry_delay != 0, wait and then retry
 ***************************************************************************/
void open_sock( int retry_delay )
{
	struct hostent *host;       /* host entry pointer */
	struct sockaddr_in sin;
	int one = 1;
	int err, status;

	if( (host = gethostbyname(Host)) == 0 ){
		sin.sin_addr.s_addr = inet_addr (Host);
		host = gethostbyaddr(  (char *)&sin.sin_addr,
			sizeof( sin.sin_addr ), AF_INET);
	}
	if( host == 0 ){
		fprintf( stderr, "%s: unknown ip-address %s\n", Name, Host );
		exit (JABORT);
	}

	memset ((char *) &sin, 0, sizeof (sin) );
	sin.sin_family = AF_INET;
	memcpy((void *)&sin.sin_addr, host->h_addr, host->h_length);
	sin.sin_family = host->h_addrtype;
	sin.sin_port = htons( Port );

	do{        /* open the socket. */
		setstatus ("connecting to %s\n", Host);
		err = errno;
		if ((sockfd = socket( AF_INET, SOCK_STREAM, 0)) < 0 ){
			fprintf( stderr, "%s: can't open socket - %s\n",
				Name, Errormsg(err) );
			exit( JFAIL );
		}
		status = connect (sockfd, (struct sockaddr *) & sin, sizeof (sin) );
		err = errno;
		if( status == 0 ){
			break;
		} else {
			fprintf( stderr, "%s: connection failed - %s\n",
				Name, Errormsg(err) );
			setstatus ("connection failed - %s\n", Errormsg(err) );
			close (sockfd);
			if( retry_delay ){
				setstatus ("sleeping %d\n", retry_delay );
				sleep(retry_delay);
			}
		}
	} while( retry_delay );                /* block forever */

#ifdef SO_KEEPALIVE
	if( sockfd >= 0 ){
		/* set SO_KEEPALIVE if it's supported */
		status = setsockopt(sockfd, SOL_SOCKET,
			SO_KEEPALIVE, (char *)&one, sizeof(one));
		err = errno;
		if( status < 0 ){
			fprintf( stderr, "%s: setsocketopt failed - %s\n",
				Name, Errormsg(err) );
		}
	}
#endif
}

int BPwrite( int fd, char *s, int cnt )
{
	int i;
	for( i = cnt;
		cnt > 0 && (i = write(fd, s, cnt)) > 0;
		cnt -= i, s += i );
	return(i);
}

