/***************************************************************************
 * LPRng - An Extended Print Spooler System
 *
 * Copyright 1988-1995 Patrick Powell, San Diego State University
 *     papowell@sdsu.edu
 * See LICENSE for conditions of use.
 *
 ***************************************************************************
 * MODULE: tcp-lp.c
 * PURPOSE: open a socket to the remote printer
 **************************************************************************/

static char _id[] = "tcp-lp.c,v 1.1 1997/01/05 04:54:10 papowell Exp";
/*
 * tcp-lp: a PLP LP-pipe command for communicating over TCP sockets.
 *
 * You are safe using the first few arguments to communicate LP-pipe-
 * specific configuration (such as the remote hostname and remote
 * port); PLP will append it's arguments to those specified in the
 * printcap entry.
 *                              -- Justin Mason 26 Aug 94.
 *
 * lppipename [-W] host port \       (these two are in the printcap entry)
 * -PPrinter -wwidth -llength -xwidth -ylength [-c] [-iindent] \
 * [-Zoptions] [-Cclass] [-Jjob] [-Raccntname] -nlogin -hHost  \
 * -Fformat -Ddebug -sstatusfile [affile]
 *                              (all added by lpd)
 */

#include "lp-pipe.h"

#define OPEN_RETRY_DELAY  10      /* delay when waiting for open */

int sockfd = -1;
int wait_til_ready;
char *Host;
int Port;
char *Name;
char *Status_file;
char buf[1024];		/* don't do write any longer than this... sigh */

void usage (void );
void tcp_pipe( void );

/***************************************************************************
 * main() - parse the various options
 * - note - we want only the -s statusfile
 *
 ***************************************************************************/
int main (int argc, char **argv )
{
	int i;
	char *s;

	/*
	 * first two args are the host/port pair.
	 */
	i = 1;
	Name = argv[0];
	if( argc <= i ) usage ();
	if( strcmp( argv[i], "-W" ) == 0 ){
		wait_til_ready = 1;
		i++;
	}
	if( argc - i < 2 ) usage ();
	Host = argv[i++];
	if( (Port = atoi( argv[i++] )) <= 0) usage ();

	/* get the -sStatusFile argument */
	while( i < argc ){
		s = argv[i++];
		if( s[0] == '-' && s[1] == 's' ){
			Status_file = s+2;
		}
	}
	(void) signal(SIGINT, kill_job);
	(void) signal(SIGPIPE, SIG_IGN);

	tcp_pipe();
	exit (JSUCC);
}

/***************************************************************************
 * tcp_pipe()
 *  - read from stdin, write to remote port
 *  - get any errors back from the remote port
 ***************************************************************************/
void tcp_pipe( void )
{
	int in, cnt;
	int status, err;		/* status and error condition */
	int eoi = 0;
	int do_exit = 0;
	fd_set rmask;
	int len;

	/* open and wait */
	open_sock( OPEN_RETRY_DELAY );

	setstatus( "printing to %s (#=1K)\n", Host);
	len = 0;
	while( !eoi ){
		FD_ZERO (&rmask);
		FD_SET (0, &rmask);     /* stdin */
		FD_SET (sockfd, &rmask);        /* socket */
		/* Find out where input is available */
		status = select( sockfd + 1, &rmask, 0, 0, 0);
		err = errno;
		/* we might have an error condition */
		if( status < 0 ){
			fprintf( stderr, "%s: select failed - %s\n",
				Name, Errormsg(err) );
			break;
		}
		if( FD_ISSET(0, &rmask) ){
			in = read(0, buf, sizeof(buf) );
			err = errno;
			switch( in ){
			case -1:        /* error */
				fprintf( stderr, "%s: read stdin failed - %s\n",
					Name, Errormsg(err) );
				do_exit = JFAIL;
				eoi = 1;
				break;
			case 0:         /* EOF */
				if( len > 0 ) setstatus("#");
				setstatus ("\nfinished sending to %s\n", Host);
				eoi = 1;
				break;
			default:
				len += in;
				cnt = BPwrite( sockfd, buf, in );
				err = errno;
				if(cnt != 0 ){
					fprintf( stderr, "%s: write sockfd failed - %s\n",
						Name, Errormsg(err) );
					do_exit = JFAIL;
					eoi = 1;
				}
				while( len >= 1024 ){
					setstatus("#");
					len -= 1024;
				}
				break;
			}
		}
		if( FD_ISSET(sockfd, &rmask) ){
			in = read(sockfd, buf, sizeof(buf) );
			err = errno;
			switch( in ){
			case -1:        /* error */
				fprintf( stderr, "%s: read socket failed - %s\n",
					Name, Errormsg(err) );
				do_exit = JFAIL;
				eoi = 1;
				break;
			case 0:         /* EOF */
				eoi = 1;
				break;
			default:
				cnt = BPwrite( 1, buf, in);
				err = errno;
				if(cnt != 0 ){
					fprintf( stderr, "%s: write stdout failed - %s\n",
						Name, Errormsg(err) );
					do_exit = JFAIL;
					eoi = 1;
				}
				break;
			}
		}
	}
	close (sockfd);

	if( wait_til_ready ){
		open_sock(OPEN_RETRY_DELAY);
		close (sockfd);
	}
	exit(do_exit);
}

/* the current job has been lprm'ed */
plp_signal_t kill_job (int signal)
{
	/* exit with a success code; we don't want the queue to stop */
	exit (JFAIL);
}

void usage (void ){
	fprintf (stderr, "usage: tcp-lp [-W] host port [... args]\n");
	exit(JABORT);
}
